@@ -38,24 +38,31 @@
 #include <telepathy-glib/dbus.h>
 #include <telepathy-glib/defs.h>
 #include <telepathy-glib/util.h>
-#include <telepathy-logger/call-event.h>
-#include <telepathy-logger/call-event-internal.h>
-#include <telepathy-logger/event-internal.h>
-#include <telepathy-logger/text-event.h>
-#include <telepathy-logger/text-event-internal.h>
-#include <telepathy-logger/log-manager.h>
-#include <telepathy-logger/log-store-internal.h>
-#include <telepathy-logger/log-manager-internal.h>
+
+#include "telepathy-logger/call-event.h"
+#include "telepathy-logger/call-event-internal.h"
+#include "telepathy-logger/entity-internal.h"
+#include "telepathy-logger/event-internal.h"
+#include "telepathy-logger/text-event.h"
+#include "telepathy-logger/text-event-internal.h"
+#include "telepathy-logger/log-manager.h"
+#include "telepathy-logger/log-store-internal.h"
+#include "telepathy-logger/log-manager-internal.h"
+#include "telepathy-logger/util-internal.h"
 
 #define DEBUG_FLAG TPL_DEBUG_LOG_STORE
-#include <telepathy-logger/entity-internal.h>
-#include <telepathy-logger/debug-internal.h>
-#include <telepathy-logger/util-internal.h>
+#include "telepathy-logger/debug-internal.h"
 
 #define LOG_DIR_CREATE_MODE       (S_IRUSR | S_IWUSR | S_IXUSR)
 #define LOG_FILE_CREATE_MODE      (S_IRUSR | S_IWUSR)
 #define LOG_DIR_CHATROOMS         "chatrooms"
 #define LOG_FILENAME_SUFFIX       ".log"
+#define LOG_FILENAME_CALL_TAG     ".call"
+#define LOG_FILENAME_CALL_SUFFIX LOG_FILENAME_CALL_TAG LOG_FILENAME_SUFFIX
+#define LOG_DATE_PATTERN          "[0-9]{8,}"
+#define LOG_FILENAME_PATTERN      "^" LOG_DATE_PATTERN "\\" LOG_FILENAME_SUFFIX "$"
+#define LOG_FILENAME_CALL_PATTERN "^" LOG_DATE_PATTERN "\\" LOG_FILENAME_CALL_TAG "\\" LOG_FILENAME_SUFFIX "$"
+
 #define LOG_TIME_FORMAT_FULL      "%Y%m%dT%H:%M:%S"
 #define LOG_TIME_FORMAT           "%Y%m%d"
 #define LOG_HEADER \
@@ -66,7 +73,12 @@
 #define LOG_FOOTER \
     "</log>\n"
 
+#ifdef ENABLE_CALL
 #define ALL_SUPPORTED_TYPES (TPL_EVENT_MASK_TEXT | TPL_EVENT_MASK_CALL)
+#else
+#define ALL_SUPPORTED_TYPES TPL_EVENT_MASK_TEXT
+#endif /* ENABLE_CALL */
+
 #define CONTAINS_ALL_SUPPORTED_TYPES(type_mask) \
   (((type_mask) & ALL_SUPPORTED_TYPES) == ALL_SUPPORTED_TYPES)
 
@@ -110,6 +122,7 @@ G_DEFINE_TYPE_WITH_CODE (TplLogStoreXml, _tpl_log_store_xml,
     G_TYPE_OBJECT,
     G_IMPLEMENT_INTERFACE (TPL_TYPE_LOG_STORE, log_store_iface_init))
 
+
 static void
 log_store_xml_dispose (GObject *object)
 {
@@ -264,7 +277,7 @@ _tpl_log_store_xml_class_init (TplLogStoreXmlClass *klass)
 
   param_spec = g_param_spec_boolean ("testmode",
       "TestMode",
-      "Wheter the logstore is in testmode, for testsuite use only",
+      "Whether the logstore is in testmode, for testsuite use only",
       FALSE, G_PARAM_WRITABLE | G_PARAM_STATIC_STRINGS);
   g_object_class_install_property (object_class, PROP_TESTMODE, param_spec);
 
@@ -313,7 +326,7 @@ log_store_xml_get_dir (TplLogStoreXml *self,
 
   escaped_account = log_store_account_to_dirname (account);
 
-  if (target)
+  if (target != NULL)
     {
       /* FIXME This may be source of bug (does that case still exist ?)
        * avoid that 1-1 conversation generated from a chatroom, having id similar
@@ -340,31 +353,45 @@ log_store_xml_get_dir (TplLogStoreXml *self,
 }
 
 
-static gchar *
-log_store_xml_get_timestamp_filename (void)
+static const gchar *
+log_store_xml_get_file_suffix (GType type)
 {
-  gchar *time_str;
+  if (type == TPL_TYPE_TEXT_EVENT)
+    return LOG_FILENAME_SUFFIX;
+  else if (type == TPL_TYPE_CALL_EVENT)
+    return LOG_FILENAME_CALL_SUFFIX;
+  else
+    g_return_val_if_reached (NULL);
+}
+
+
+static gchar *
+log_store_xml_get_timestamp_filename (GType type,
+    gint64 timestamp)
+{
+  gchar *date_str;
   gchar *filename;
-  GDateTime *now;
+  GDateTime *date;
 
-  now = g_date_time_new_now_local ();
-  time_str = g_date_time_format (now, LOG_TIME_FORMAT);
-  filename = g_strconcat (time_str, LOG_FILENAME_SUFFIX, NULL);
+  date = g_date_time_new_from_unix_utc (timestamp);
+  date_str = g_date_time_format (date, LOG_TIME_FORMAT);
+  filename = g_strconcat (date_str, log_store_xml_get_file_suffix (type),
+      NULL);
 
-  g_date_time_unref (now);
-  g_free (time_str);
+  g_date_time_unref (date);
+  g_free (date_str);
 
   return filename;
 }
 
 
 static gchar *
-log_store_xml_get_timestamp_from_event (TplEvent *event)
+log_store_xml_format_timestamp (gint64 timestamp)
 {
   GDateTime *ts;
   gchar *ts_str;
 
-  ts = g_date_time_new_from_unix_utc (tpl_event_get_timestamp (event));
+  ts = g_date_time_new_from_unix_utc (timestamp);
   ts_str = g_date_time_format (ts, LOG_TIME_FORMAT_FULL);
 
   g_date_time_unref (ts);
@@ -373,21 +400,30 @@ log_store_xml_get_timestamp_from_event (TplEvent *event)
 }
 
 
+static gchar *
+log_store_xml_get_timestamp_from_event (TplEvent *event)
+{
+  return log_store_xml_format_timestamp (tpl_event_get_timestamp (event));
+}
+
+
 static gchar *
 log_store_xml_get_filename (TplLogStoreXml *self,
     TpAccount *account,
-    TplEntity *target)
+    TplEntity *target,
+    GType type,
+    gint64 timestamp)
 {
   gchar *id_dir;
-  gchar *timestamp;
+  gchar *timestamp_str;
   gchar *filename;
 
   id_dir = log_store_xml_get_dir (self, account, target);
-  timestamp = log_store_xml_get_timestamp_filename ();
-  filename = g_build_filename (id_dir, timestamp, NULL);
+  timestamp_str = log_store_xml_get_timestamp_filename (type, timestamp);
+  filename = g_build_filename (id_dir, timestamp_str, NULL);
 
   g_free (id_dir);
-  g_free (timestamp);
+  g_free (timestamp_str);
 
   return filename;
 }
@@ -401,6 +437,8 @@ _log_store_xml_write_to_store (TplLogStoreXml *self,
     TpAccount *account,
     TplEntity *target,
     const gchar *event,
+    GType type,
+    gint64 timestamp,
     GError **error)
 {
   FILE *file;
@@ -413,7 +451,8 @@ _log_store_xml_write_to_store (TplLogStoreXml *self,
   g_return_val_if_fail (TP_IS_ACCOUNT (account), FALSE);
   g_return_val_if_fail (TPL_IS_ENTITY (target), FALSE);
 
-  filename = log_store_xml_get_filename (self, account, target);
+
+  filename = log_store_xml_get_filename (self, account, target, type, timestamp);
   basedir = g_path_get_dirname (filename);
 
   if (!g_file_test (basedir, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR))
@@ -464,16 +503,18 @@ add_text_event (TplLogStoreXml *self,
     GError **error)
 {
   gboolean ret = FALSE;
+  gint64 timestamp;
   TpDBusDaemon *bus_daemon;
   TpAccount *account;
   TplEntity *sender;
   const gchar *body_str;
+  const gchar *token_str;
   gchar *avatar_token = NULL;
   gchar *body = NULL;
-  gchar *timestamp = NULL;
+  gchar *time_str = NULL;
   gchar *contact_name = NULL;
   gchar *contact_id = NULL;
-  gchar *event = NULL;
+  GString *event = NULL;
   TpChannelTextMessageType msg_type;
 
   g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
@@ -500,7 +541,7 @@ add_text_event (TplLogStoreXml *self,
 
   body = g_markup_escape_text (body_str, -1);
   msg_type = tpl_text_event_get_message_type (message);
-  timestamp = log_store_xml_get_timestamp_from_event (
+  time_str = log_store_xml_get_timestamp_from_event (
       TPL_EVENT (message));
 
   sender = tpl_event_get_sender (TPL_EVENT (message));
@@ -509,28 +550,60 @@ add_text_event (TplLogStoreXml *self,
   avatar_token = g_markup_escape_text (tpl_entity_get_avatar_token (sender),
       -1);
 
-  event = g_strdup_printf ("<message time='%s' id='%s' name='%s' "
-      "token='%s' isuser='%s' type='%s'>"
-      "%s</message>\n" LOG_FOOTER, timestamp,
-      contact_id, contact_name,
+  event = g_string_new (NULL);
+  g_string_printf (event, "<message time='%s' id='%s' name='%s' "
+      "token='%s' isuser='%s' type='%s'",
+      time_str, contact_id, contact_name,
       avatar_token,
       tpl_entity_get_entity_type (sender)
           == TPL_ENTITY_SELF ? "true" : "false",
-      _tpl_text_event_message_type_to_str (msg_type),
-      body);
+      _tpl_text_event_message_type_to_str (msg_type));
+
+  token_str = tpl_text_event_get_message_token (message);
+  if (!TPL_STR_EMPTY (token_str))
+    {
+      gchar *message_token = g_markup_escape_text (token_str, -1);
+      g_string_append_printf (event, " message-token='%s'", message_token);
+      g_free (message_token);
+
+      token_str = tpl_text_event_get_supersedes_token (message);
+      if (!TPL_STR_EMPTY (token_str))
+        {
+          gchar *supersedes_token = g_markup_escape_text (token_str, -1);
+          guint edit_timestamp;
+          g_string_append_printf (event, " supersedes-token='%s'",
+              supersedes_token);
+
+          edit_timestamp = tpl_text_event_get_edit_timestamp (message);
+          if (edit_timestamp != 0)
+            {
+              gchar *edit_timestamp_str =
+                  log_store_xml_format_timestamp (edit_timestamp);
+              g_string_append_printf (event, " edit-timestamp='%s'",
+                  edit_timestamp_str);
+              g_free (edit_timestamp_str);
+            }
+        }
+
+    }
+
+  timestamp = tpl_event_get_timestamp (TPL_EVENT (message));
+
+  g_string_append_printf (event, ">%s</message>\n" LOG_FOOTER, body);
 
   DEBUG ("writing text event from %s (ts %s)",
-      contact_id, timestamp);
+      contact_id, time_str);
 
   ret = _log_store_xml_write_to_store (self, account,
-      _tpl_event_get_target (TPL_EVENT (message)), event, error);
+      _tpl_event_get_target (TPL_EVENT (message)), event->str,
+      TPL_TYPE_TEXT_EVENT, timestamp, error);
 
 out:
   g_free (contact_id);
   g_free (contact_name);
-  g_free (timestamp);
+  g_free (time_str);
   g_free (body);
-  g_free (event);
+  g_string_free (event, TRUE);
   g_free (avatar_token);
 
   if (bus_daemon != NULL)
@@ -539,6 +612,7 @@ out:
   return ret;
 }
 
+
 static gboolean
 add_call_event (TplLogStoreXml *self,
     TplCallEvent *event,
@@ -550,14 +624,14 @@ add_call_event (TplLogStoreXml *self,
   TplEntity *sender;
   TplEntity *actor;
   TplEntity *target;
-  gchar *timestamp;
+  gchar *time_str = NULL;
   gchar *sender_avatar = NULL;
   gchar *sender_name = NULL;
-  gchar *sender_id;
+  gchar *sender_id = NULL;
   gchar *actor_name = NULL;
   gchar *actor_avatar = NULL;
-  gchar *actor_id;
-  gchar *log_str;
+  gchar *actor_id = NULL;
+  gchar *log_str = NULL;
   TplCallEndReason reason;
 
   g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
@@ -573,7 +647,7 @@ add_call_event (TplLogStoreXml *self,
 
   account = tpl_event_get_account (TPL_EVENT (event));
 
-  timestamp = log_store_xml_get_timestamp_from_event (
+  time_str = log_store_xml_get_timestamp_from_event (
       TPL_EVENT (event));
   reason = tpl_call_event_get_end_reason (event);
 
@@ -597,7 +671,7 @@ add_call_event (TplLogStoreXml *self,
       "actorname='%s' actortoken='%s' "
       "reason='%s' detail='%s'/>\n"
       LOG_FOOTER,
-        timestamp,
+        time_str,
         sender_id,
         sender_name,
         tpl_entity_get_entity_type (sender) ==
@@ -613,9 +687,11 @@ add_call_event (TplLogStoreXml *self,
 
   DEBUG ("writing call event from %s (ts %s)",
       tpl_entity_get_identifier (target),
-      timestamp);
+      time_str);
 
-  ret = _log_store_xml_write_to_store (self, account, target, log_str, error);
+  ret = _log_store_xml_write_to_store (self, account, target, log_str,
+      TPL_TYPE_CALL_EVENT, tpl_event_get_timestamp (TPL_EVENT (event)),
+      error);
 
 out:
   g_free (sender_id);
@@ -624,7 +700,7 @@ out:
   g_free (actor_id);
   g_free (actor_name);
   g_free (actor_avatar);
-  g_free (timestamp);
+  g_free (time_str);
   g_free (log_str);
 
   if (bus_daemon != NULL)
@@ -658,6 +734,91 @@ log_store_xml_add_event (TplLogStore *store,
 }
 
 
+static gboolean
+log_store_xml_exists_in_directory (const gchar *dirname,
+    GRegex *regex,
+    gint type_mask,
+    gboolean recursive)
+{
+  gboolean exists;
+  const gchar *basename;
+  GDir *dir;
+
+  DEBUG ("Looking in directory '%s' %s",
+      dirname, recursive ? "resursively" : "");
+
+  dir = g_dir_open (dirname, 0, NULL);
+  exists = (dir != NULL);
+
+  if (CONTAINS_ALL_SUPPORTED_TYPES (type_mask) || !exists)
+    goto out;
+
+  exists = FALSE;
+  while ((basename = g_dir_read_name (dir)) != NULL)
+    {
+      gchar *filename;
+
+      filename = g_build_filename (dirname, basename, NULL);
+
+      DEBUG ("Matching with filename '%s'", basename);
+
+      if (recursive && g_file_test (filename, G_FILE_TEST_IS_DIR))
+        exists = log_store_xml_exists_in_directory (filename, regex, type_mask,
+            !tp_strdiff (basename, LOG_DIR_CHATROOMS));
+      else if (g_file_test (filename, G_FILE_TEST_IS_REGULAR))
+        exists = g_regex_match (regex, basename, 0, 0);
+
+      g_free (filename);
+
+      if (exists)
+        break;
+    }
+
+out:
+  if (dir != NULL)
+    g_dir_close (dir);
+
+  return exists;
+}
+
+
+static GRegex *
+log_store_xml_create_filename_regex (gint type_mask)
+{
+  GString *pattern;
+  GRegex *regex = NULL;
+  GError *error = NULL;
+
+  pattern = g_string_new ("");
+
+  if (type_mask & TPL_EVENT_MASK_TEXT)
+    g_string_append (pattern, LOG_FILENAME_PATTERN);
+
+  if (type_mask & TPL_EVENT_MASK_CALL)
+    g_string_append_printf (pattern,
+        "%s" LOG_FILENAME_CALL_PATTERN,
+        pattern->len == 0 ? "" : "|");
+
+  if (pattern->len == 0)
+    goto out;
+
+  DEBUG ("Pattern is '%s'", pattern->str);
+
+  regex = g_regex_new (pattern->str, G_REGEX_OPTIMIZE, 0, &error);
+
+  if (regex == NULL)
+    {
+      DEBUG ("Failed to create regex: %s", error->message);
+      g_error_free (error);
+    }
+
+out:
+  g_string_free (pattern, TRUE);
+
+  return regex;
+}
+
+
 static gboolean
 log_store_xml_exists (TplLogStore *store,
     TpAccount *account,
@@ -665,20 +826,25 @@ log_store_xml_exists (TplLogStore *store,
     gint type_mask)
 {
   TplLogStoreXml *self = (TplLogStoreXml *) store;
-  gchar *dir;
-  gboolean exists;
+  gchar *dirname;
+  GRegex *regex;
+  gboolean exists = FALSE;
 
   g_return_val_if_fail (TPL_IS_LOG_STORE_XML (self), FALSE);
   g_return_val_if_fail (TP_IS_ACCOUNT (account), FALSE);
-  g_return_val_if_fail (TPL_IS_ENTITY (target), FALSE);
+  g_return_val_if_fail (target == NULL || TPL_IS_ENTITY (target), FALSE);
 
-  /* FIXME This method is exposed synchronously in the log manager API and
-   * thus we need a constant time reply. The implementation is not 100%
-   * correct here, but provide this constant time. See fd.o but #35549. */
+  dirname = log_store_xml_get_dir (self, account, target);
+  regex = log_store_xml_create_filename_regex (type_mask);
 
-  dir = log_store_xml_get_dir (self, account, target);
-  exists = g_file_test (dir, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR);
-  g_free (dir);
+  if (regex != NULL)
+    exists = log_store_xml_exists_in_directory (dirname, regex, type_mask,
+        target == NULL);
+
+  g_free (dirname);
+
+  if (regex != NULL)
+    g_regex_unref (regex);
 
   return exists;
 }
@@ -752,7 +918,6 @@ log_store_xml_get_dates (TplLogStore *store,
   GDir *dir = NULL;
   GString *pattern = NULL;
   GRegex *regex = NULL;
-  GError *error = NULL;
   const gchar *basename;
 
   g_return_val_if_fail (TPL_IS_LOG_STORE_XML (self), NULL);
@@ -768,59 +933,37 @@ log_store_xml_get_dates (TplLogStore *store,
     }
 
   DEBUG ("Collating a list of dates in:'%s'", directory);
+  regex = log_store_xml_create_filename_regex (type_mask);
 
-  pattern = g_string_new ("");
-
-  if (type_mask & TPL_EVENT_MASK_TEXT)
-    g_string_append (pattern, "<message ");
-
-  if (type_mask & TPL_EVENT_MASK_CALL)
-    g_string_append_printf (pattern,
-        "%s<call ",
-        pattern->len == 0 ? "" : "|");
-
-  if (pattern->len == 0)
-    goto out;
-
-  regex = g_regex_new (pattern->str, G_REGEX_OPTIMIZE, 0, &error);
   if (regex == NULL)
-    {
-      DEBUG ("Failed to create regex: %s", error->message);
-      g_error_free (error);
-      goto out;
-    }
+    goto out;
 
   while ((basename = g_dir_read_name (dir)) != NULL)
     {
-      gchar *filename;
+      const gchar *p;
+      gchar *str;
+      GDate *date;
 
-      if (!g_str_has_suffix (basename, LOG_FILENAME_SUFFIX))
+      if (!g_regex_match (regex, basename, 0, NULL))
         continue;
 
-      filename = g_build_filename (directory, basename, NULL);
+      p = strstr (basename, LOG_FILENAME_CALL_SUFFIX);
 
-      if (CONTAINS_ALL_SUPPORTED_TYPES (type_mask)
-          || log_store_xml_match_in_file (filename, regex))
-        {
-          const gchar *p;
-          gchar *str;
-          GDate *date;
+      if (p == NULL)
+        p = strstr (basename, LOG_FILENAME_SUFFIX);
 
-          p = strstr (basename, LOG_FILENAME_SUFFIX);
-          str = g_strndup (basename, p - basename);
+      str = g_strndup (basename, p - basename);
 
-          if (str == NULL)
-            continue;
+      if (str == NULL)
+        continue;
 
-          date = create_date_from_string (str);
-          if (date != NULL)
-            dates = g_list_insert_sorted (dates, date,
-                (GCompareFunc) g_date_compare);
+      date = create_date_from_string (str);
 
-          g_free (str);
-        }
+      if (date != NULL)
+        dates = g_list_insert_sorted (dates, date,
+            (GCompareFunc) g_date_compare);
 
-      g_free (filename);
+      g_free (str);
     }
 
 out:
@@ -845,7 +988,8 @@ static gchar *
 log_store_xml_get_filename_for_date (TplLogStoreXml *self,
     TpAccount *account,
     TplEntity *target,
-    const GDate *date)
+    const GDate *date,
+    GType type)
 {
   gchar *basedir;
   gchar *timestamp;
@@ -860,7 +1004,7 @@ log_store_xml_get_filename_for_date (TplLogStoreXml *self,
   g_date_strftime (str, 9, "%Y%m%d", date);
 
   basedir = log_store_xml_get_dir (self, account, target);
-  timestamp = g_strconcat (str, LOG_FILENAME_SUFFIX, NULL);
+  timestamp = g_strconcat (str, log_store_xml_get_file_suffix (type), NULL);
   filename = g_build_filename (basedir, timestamp, NULL);
 
   g_free (basedir);
@@ -952,10 +1096,14 @@ parse_text_node (TplLogStoreXml *self,
   TplEntity *receiver;
   gchar *time_str;
   gint64 timestamp;
+  gchar *edit_time_str;
+  gint64 edit_timestamp = 0;
   gchar *sender_id;
   gchar *sender_name;
   gchar *sender_avatar_token;
   gchar *body;
+  gchar *message_token;
+  gchar *supersedes_token;
   gchar *is_user_str;
   gboolean is_user = FALSE;
   gchar *msg_type_str;
@@ -963,10 +1111,16 @@ parse_text_node (TplLogStoreXml *self,
 
   body = (gchar *) xmlNodeGetContent (node);
   time_str = (gchar *) xmlGetProp (node, (const xmlChar *) "time");
+  edit_time_str = (gchar *) xmlGetProp (node,
+      (const xmlChar *) "edit-timestamp");
   sender_id = (gchar *) xmlGetProp (node, (const xmlChar *) "id");
   sender_name = (gchar *) xmlGetProp (node, (const xmlChar *) "name");
   sender_avatar_token = (gchar *) xmlGetProp (node,
       (const xmlChar *) "token");
+  message_token = (gchar *) xmlGetProp (node,
+      (const xmlChar *) "message-token");
+  supersedes_token = (gchar *) xmlGetProp (node,
+      (const xmlChar *) "supersedes-token");
   is_user_str = (gchar *) xmlGetProp (node, (const xmlChar *) "isuser");
   msg_type_str = (gchar *) xmlGetProp (node, (const xmlChar *) "type");
 
@@ -978,6 +1132,11 @@ parse_text_node (TplLogStoreXml *self,
 
   timestamp = _tpl_time_parse (time_str);
 
+  if (supersedes_token != NULL && edit_time_str != NULL)
+    {
+      edit_timestamp = _tpl_time_parse (edit_time_str);
+    }
+
   if (is_room)
     receiver = tpl_entity_new_from_room_id (target_id);
   else if (is_user)
@@ -1000,14 +1159,20 @@ parse_text_node (TplLogStoreXml *self,
       /* TplTextEvent */
       "message-type", msg_type,
       "message", body,
+      "message-token", message_token,
+      "supersedes-token", supersedes_token,
+      "edit-timestamp", edit_timestamp,
       NULL);
 
   g_object_unref (sender);
   g_object_unref (receiver);
   xmlFree (time_str);
+  xmlFree (edit_time_str);
   xmlFree (sender_id);
   xmlFree (sender_name);
   xmlFree (body);
+  xmlFree (message_token);
+  xmlFree (supersedes_token);
   xmlFree (is_user_str);
   xmlFree (msg_type_str);
   xmlFree (sender_avatar_token);
@@ -1118,14 +1283,86 @@ parse_call_node (TplLogStoreXml *self,
   return event;
 }
 
-/* returns a Glist of TplEvent instances */
+
+static void
+event_queue_replace_and_supersede (GQueue *events,
+    GList *index,
+    GHashTable *superseded_links,
+    TplTextEvent *event)
+{
+  _tpl_text_event_add_supersedes (event, index->data);
+  g_hash_table_insert (superseded_links,
+      (gpointer) tpl_text_event_get_message_token (index->data), index);
+  g_object_unref (index->data);
+  index->data = event;
+}
+
+
 static GList *
+event_queue_add_text_event (GQueue *events,
+    GList *index,
+    GHashTable *superseded_links,
+    TplTextEvent *event)
+{
+  GList *l = NULL;
+  const gchar *supersedes_token = tpl_text_event_get_supersedes_token (event);
+  TplTextEvent *dummy_event;
+
+  if (supersedes_token == NULL)
+    return _tpl_event_queue_insert_sorted_after (events, index,
+        TPL_EVENT (event));
+
+  l = g_hash_table_lookup (superseded_links, supersedes_token);
+  if (l != NULL)
+    {
+      event_queue_replace_and_supersede (events, l, superseded_links, event);
+      return index;
+    }
+
+  /* Search backwards from "now" and insert (but don't update "now") */
+  for (l = index; l != NULL; l = g_list_previous (l))
+    {
+      if (!tp_strdiff (tpl_text_event_get_message_token (l->data),
+          supersedes_token))
+        {
+          event_queue_replace_and_supersede (events, l, superseded_links,
+              event);
+          return index;
+        }
+    }
+
+  DEBUG ("Can't find event %s (superseded by %s). "
+      "Adding Dummy event.",
+      supersedes_token, tpl_text_event_get_message_token (event));
+
+  dummy_event = g_object_new (TPL_TYPE_TEXT_EVENT,
+      /* TplEvent */
+      "account", tpl_event_get_account (TPL_EVENT (event)),
+      /* MISSING: "channel-path", channel_path, */
+      "receiver", tpl_event_get_receiver (TPL_EVENT (event)),
+      "sender", tpl_event_get_sender (TPL_EVENT (event)),
+      "timestamp", tpl_event_get_timestamp (TPL_EVENT (event)),
+      /* TplTextEvent */
+      "message-type", tpl_text_event_get_message_type (event),
+      "message", "",
+      "message-token", supersedes_token,
+      NULL);
+
+  index = _tpl_event_queue_insert_sorted_after (events, index,
+      TPL_EVENT (dummy_event));
+  event_queue_replace_and_supersede (events, index, superseded_links, event);
+  return index;
+}
+
+
+/* returns a Glist of TplEvent instances */
+static void
 log_store_xml_get_events_for_file (TplLogStoreXml *self,
     TpAccount *account,
     const gchar *filename,
-    gint type_mask)
+    GType type,
+    GQueue *events)
 {
-  GQueue events = { 0 };
   xmlParserCtxtPtr ctxt;
   xmlDocPtr doc;
   xmlNodePtr log_node;
@@ -1135,18 +1372,21 @@ log_store_xml_get_events_for_file (TplLogStoreXml *self,
   gchar *tmp;
   gchar *target_id;
   gchar *self_id;
+  GHashTable *supersedes_links;
   GError *error = NULL;
+  guint num_events = 0;
+  GList *index;
 
-  g_return_val_if_fail (TPL_IS_LOG_STORE_XML (self), NULL);
-  g_return_val_if_fail (TP_IS_ACCOUNT (account), NULL);
-  g_return_val_if_fail (!TPL_STR_EMPTY (filename), NULL);
+  g_return_if_fail (TPL_IS_LOG_STORE_XML (self));
+  g_return_if_fail (TP_IS_ACCOUNT (account));
+  g_return_if_fail (!TPL_STR_EMPTY (filename));
 
   DEBUG ("Attempting to parse filename:'%s'...", filename);
 
   if (!g_file_test (filename, G_FILE_TEST_EXISTS))
     {
       DEBUG ("Filename:'%s' does not exist", filename);
-      return NULL;
+      return;
     }
 
   if (!tp_account_parse_object_path (
@@ -1156,7 +1396,7 @@ log_store_xml_get_events_for_file (TplLogStoreXml *self,
       DEBUG ("Cannot get self identifier from account: %s",
           error->message);
       g_error_free (error);
-      return NULL;
+      return;
     }
 
   /* Create parser. */
@@ -1169,7 +1409,7 @@ log_store_xml_get_events_for_file (TplLogStoreXml *self,
       g_warning ("Failed to parse file:'%s'", filename);
       xmlFreeParserCtxt (ctxt);
       g_free (self_id);
-      return NULL;
+      return;
     }
 
   /* The root node, presets. */
@@ -1179,7 +1419,7 @@ log_store_xml_get_events_for_file (TplLogStoreXml *self,
       xmlFreeDoc (doc);
       xmlFreeParserCtxt (ctxt);
       g_free (self_id);
-      return NULL;
+      return;
     }
 
   /* Guess the target based on directory name */
@@ -1195,31 +1435,50 @@ log_store_xml_get_events_for_file (TplLogStoreXml *self,
   g_free (dirname);
   g_free (tmp);
 
+  /* Temporary hash from (borrowed) supersedes-token to (borrowed) link in
+   * events, for any event that was once in events, but has since been
+   * superseded (and therefore won't be found by a linear search). */
+  supersedes_links = g_hash_table_new (g_str_hash, g_str_equal);
+
   /* Now get the events. */
+  index = NULL;
   for (node = log_node->children; node; node = node->next)
     {
       TplEvent *event = NULL;
 
-      if (type_mask & TPL_EVENT_MASK_TEXT
+      if (type == TPL_TYPE_TEXT_EVENT
           && strcmp ((const gchar *) node->name, "message") == 0)
-        event = parse_text_node (self, node, is_room, target_id, self_id,
-            account);
-      else if (type_mask & TPL_EVENT_MASK_CALL
-          && strcmp ((const char*) node->name, "call") == 0)
-        event = parse_call_node (self, node, is_room, target_id, self_id,
-            account);
+        {
+          event = parse_text_node (self, node, is_room, target_id, self_id,
+              account);
 
-      if (event != NULL)
-        g_queue_push_tail (&events, event);
+          if (event == NULL)
+            continue;
+
+          index = event_queue_add_text_event (events, index,
+              supersedes_links, TPL_TEXT_EVENT (event));
+          num_events++;
+        }
+      else if (type == TPL_TYPE_CALL_EVENT
+          && strcmp ((const char*) node->name, "call") == 0)
+        {
+          event = parse_call_node (self, node, is_room, target_id, self_id,
+              account);
+
+          if (event == NULL)
+            continue;
+
+          index = _tpl_event_queue_insert_sorted_after (events, index, event);
+          num_events++;
+        }
     }
 
-  DEBUG ("Parsed %d events", events.length);
+  DEBUG ("Parsed %u events", num_events);
 
   g_free (target_id);
   xmlFreeDoc (doc);
   xmlFreeParserCtxt (ctxt);
-
-  return events.head;
+  g_hash_table_destroy (supersedes_links);
 }
 
 
@@ -1227,12 +1486,14 @@ log_store_xml_get_events_for_file (TplLogStoreXml *self,
  * Used to make possible the full search vs. specific subtrees search */
 static GList *
 log_store_xml_get_all_files (TplLogStoreXml *self,
-    const gchar *dir)
+    const gchar *dir,
+    gint type_mask)
 {
   GDir *gdir;
   GList *files = NULL;
   const gchar *name;
   const gchar *basedir;
+  GRegex *regex;
 
   g_return_val_if_fail (TPL_IS_LOG_STORE_XML (self), NULL);
   /* dir can be NULL, do not check */
@@ -1243,30 +1504,34 @@ log_store_xml_get_all_files (TplLogStoreXml *self,
   if (!gdir)
     return NULL;
 
+  regex = log_store_xml_create_filename_regex (type_mask);
+
+  if (regex == NULL)
+    goto out;
+
   while ((name = g_dir_read_name (gdir)) != NULL)
     {
       gchar *filename;
 
       filename = g_build_filename (basedir, name, NULL);
-      if (g_str_has_suffix (filename, LOG_FILENAME_SUFFIX))
-        {
-          files = g_list_prepend (files, filename);
-          continue;
-        }
 
-      if (g_file_test (filename, G_FILE_TEST_IS_DIR))
+      if (g_regex_match (regex, name, 0, NULL))
+        files = g_list_prepend (files, filename);
+      else if (g_file_test (filename, G_FILE_TEST_IS_DIR))
         {
           /* Recursively get all log files */
           files = g_list_concat (files,
-              log_store_xml_get_all_files (self,
-                filename));
+              log_store_xml_get_all_files (self, filename, type_mask));
+          g_free (filename);
         }
-
-      g_free (filename);
     }
 
+out:
   g_dir_close (gdir);
 
+  if (regex != NULL)
+    g_regex_unref (regex);
+
   return files;
 }
 
@@ -1369,12 +1634,13 @@ log_store_xml_search_new (TplLogStore *store,
   g_return_val_if_fail (TPL_IS_LOG_STORE_XML (self), NULL);
   g_return_val_if_fail (!TPL_STR_EMPTY (text), NULL);
 
-  files = log_store_xml_get_all_files (self, NULL);
+  files = log_store_xml_get_all_files (self, NULL, type_mask);
   DEBUG ("Found %d log files in total", g_list_length (files));
 
   return _log_store_xml_search_in_files (self, text, files, type_mask);
 }
 
+
 /* Returns: (GList *) of (TplLogSearchHit *) */
 static GList *
 log_store_xml_get_entities_for_dir (TplLogStoreXml *self,
@@ -1435,19 +1701,32 @@ log_store_xml_get_events_for_date (TplLogStore *store,
 {
   TplLogStoreXml *self = (TplLogStoreXml *) store;
   gchar *filename;
-  GList *events;
+  GQueue events = G_QUEUE_INIT;
 
   g_return_val_if_fail (TPL_IS_LOG_STORE_XML (self), NULL);
   g_return_val_if_fail (TP_IS_ACCOUNT (account), NULL);
   g_return_val_if_fail (TPL_IS_ENTITY (target), NULL);
   g_return_val_if_fail (date != NULL, NULL);
 
-  filename = log_store_xml_get_filename_for_date (self, account, target,
-      date);
-  events = log_store_xml_get_events_for_file (self, account, filename, type_mask);
-  g_free (filename);
+  if (type_mask & TPL_EVENT_MASK_TEXT)
+    {
+      filename = log_store_xml_get_filename_for_date (self, account, target,
+          date, TPL_TYPE_TEXT_EVENT);
+      log_store_xml_get_events_for_file (self, account, filename,
+          TPL_TYPE_TEXT_EVENT, &events);
+      g_free (filename);
+    }
 
-  return events;
+  if (type_mask & TPL_EVENT_MASK_CALL)
+    {
+      filename = log_store_xml_get_filename_for_date (self, account, target,
+          date, TPL_TYPE_CALL_EVENT);
+      log_store_xml_get_events_for_file (self, account, filename,
+          TPL_TYPE_CALL_EVENT, &events);
+      g_free (filename);
+    }
+
+  return events.head;
 }
 
 
