@@ -1,4 +1,3 @@
-/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
  * Copyright (C) 2009 Collabora Ltd.
  *
@@ -380,6 +379,35 @@ _tpl_observer_init (TplObserver *self)
            TP_HANDLE_TYPE_ROOM,
       NULL));
 
+#ifdef ENABLE_CALL
+  /* Observe contact stream media channels */
+  tp_base_client_take_observer_filter (TP_BASE_CLIENT (self),
+      tp_asv_new (
+       TP_PROP_CHANNEL_CHANNEL_TYPE, G_TYPE_STRING,
+           TP_IFACE_CHANNEL_TYPE_STREAMED_MEDIA,
+       TP_PROP_CHANNEL_TARGET_HANDLE_TYPE, G_TYPE_UINT,
+           TP_HANDLE_TYPE_CONTACT,
+      NULL));
+
+  /* Observe contact call channels */
+  tp_base_client_take_observer_filter (TP_BASE_CLIENT (self),
+      tp_asv_new (
+       TP_PROP_CHANNEL_CHANNEL_TYPE, G_TYPE_STRING,
+           "org.freedesktop.Telepathy.Channel.Type.Call.DRAFT",
+       TP_PROP_CHANNEL_TARGET_HANDLE_TYPE, G_TYPE_UINT,
+           TP_HANDLE_TYPE_CONTACT,
+      NULL));
+
+  /* Observe room call channels */
+  tp_base_client_take_observer_filter (TP_BASE_CLIENT (self),
+      tp_asv_new (
+       TP_PROP_CHANNEL_CHANNEL_TYPE, G_TYPE_STRING,
+           "org.freedesktop.Telepathy.Channel.Type.Call.DRAFT",
+       TP_PROP_CHANNEL_TARGET_HANDLE_TYPE, G_TYPE_UINT,
+           TP_HANDLE_TYPE_ROOM,
+      NULL));
+#endif /* ENABLE_CALL */
+
   tp_base_client_set_observer_recover (TP_BASE_CLIENT (self), TRUE);
 }
 
