@@ -216,14 +216,16 @@ get_remote_contact_cb (TpConnection *connection,
       GError *new_error = NULL;
       new_error = g_error_new (error->domain, error->code,
           "Failed to get remote contact: %s", error->message);
-      _tpl_action_chain_terminate (ctx, error);
+      _tpl_action_chain_terminate (ctx, new_error);
+      g_error_free (new_error);
     }
   else if (n_failed > 0)
     {
-      error = g_error_new (TPL_TEXT_CHANNEL_ERROR,
+      GError *new_error = g_error_new (TPL_TEXT_CHANNEL_ERROR,
           TPL_TEXT_CHANNEL_ERROR_FAILED,
           "Failed to prepare remote contact.");
-      _tpl_action_chain_terminate (ctx, error);
+      _tpl_action_chain_terminate (ctx, new_error);
+      g_error_free (new_error);
     }
   else
     {
@@ -305,11 +307,29 @@ get_message_pending_id (TpMessage *m)
 }
 
 
-static guint
-get_message_timestamp (TpMessage *message)
+static gint64
+get_original_message_timestamp (TpMessage *message)
 {
   gint64 timestamp;
 
+  timestamp = tp_asv_get_int64 (tp_message_peek (message, 0),
+      "original-message-sent", NULL);
+
+  if (timestamp == 0)
+    timestamp = tp_asv_get_int64 (tp_message_peek (message, 0),
+        "original-message-received", NULL);
+
+  return timestamp;
+}
+
+
+static gint64
+get_network_timestamp (TpMessage *message)
+{
+  GDateTime *datetime = g_date_time_new_now_utc ();
+  gint64 now = g_date_time_to_unix (datetime);
+  gint64 timestamp;
+
   timestamp = tp_message_get_sent_timestamp (message);
 
   if (timestamp == 0)
@@ -317,11 +337,41 @@ get_message_timestamp (TpMessage *message)
 
   if (timestamp == 0)
     {
-      GDateTime *datetime = g_date_time_new_now_utc ();
-      timestamp = g_date_time_to_unix (datetime);
-      g_date_time_unref (datetime);
+      DEBUG ("TpMessage is not timestamped. Using current time instead.");
+      timestamp = now;
     }
 
+  if (timestamp - now > 60 * 60)
+    DEBUG ("timestamp is more than an hour in the future.");
+  else  if (now - timestamp > 60 * 60)
+    DEBUG ("timestamp is more than an hour in the past.");
+
+  g_date_time_unref (datetime);
+
+  return timestamp;
+}
+
+
+static gint64
+get_message_edit_timestamp (TpMessage *message)
+{
+  if (tp_message_get_supersedes (message) != NULL)
+    return get_network_timestamp (message);
+  else
+    return 0;
+}
+
+
+static gint64
+get_message_timestamp (TpMessage *message)
+{
+  gint64 timestamp;
+
+  timestamp = get_original_message_timestamp (message);
+
+  if (timestamp == 0)
+    timestamp = get_network_timestamp (message);
+
   return timestamp;
 }
 
@@ -395,6 +445,9 @@ tpl_text_channel_store_message (TplTextChannel *self,
       "receiver", receiver,
       "sender", sender,
       "timestamp", timestamp,
+      "message-token", tp_message_get_token (message),
+      "supersedes-token", tp_message_get_supersedes (message),
+      "edit-timestamp", get_message_edit_timestamp (message),
       /* TplTextEvent */
       "message-type", type,
       "message", text,
