@@ -52,13 +52,22 @@ G_DEFINE_TYPE (TplTextEvent, tpl_text_event, TPL_TYPE_EVENT)
 struct _TplTextEventPriv
 {
   TpChannelTextMessageType message_type;
+  gint64 edit_timestamp;
   gchar *message;
+  gchar *token;
+  gchar *supersedes_token;
+  /* A list of TplTextEvent that we supersede.
+   * This is only populated when reading logs (not storing them). */
+  GQueue supersedes;
 };
 
 enum
 {
   PROP_MESSAGE_TYPE = 1,
-  PROP_MESSAGE
+  PROP_EDIT_TIMESTAMP,
+  PROP_MESSAGE,
+  PROP_TOKEN,
+  PROP_SUPERSEDES
 };
 
 static gchar *message_types[] = {
@@ -71,6 +80,17 @@ static gchar *message_types[] = {
 };
 
 
+static void
+tpl_text_event_dispose (GObject *obj)
+{
+  TplTextEventPriv *priv = TPL_TEXT_EVENT (obj)->priv;
+
+  g_list_foreach (priv->supersedes.head, (GFunc) g_object_unref, NULL);
+  g_list_free (priv->supersedes.head);
+  g_queue_init (&priv->supersedes);
+}
+
+
 static void
 tpl_text_event_finalize (GObject *obj)
 {
@@ -79,6 +99,12 @@ tpl_text_event_finalize (GObject *obj)
   g_free (priv->message);
   priv->message = NULL;
 
+  g_free (priv->token);
+  priv->token = NULL;
+
+  g_free (priv->supersedes_token);
+  priv->supersedes_token = NULL;
+
   G_OBJECT_CLASS (tpl_text_event_parent_class)->finalize (obj);
 }
 
@@ -96,9 +122,18 @@ tpl_text_event_get_property (GObject *object,
       case PROP_MESSAGE_TYPE:
         g_value_set_uint (value, priv->message_type);
         break;
+      case PROP_EDIT_TIMESTAMP:
+        g_value_set_int64 (value, priv->edit_timestamp);
+        break;
       case PROP_MESSAGE:
         g_value_set_string (value, priv->message);
         break;
+      case PROP_TOKEN:
+        g_value_set_string (value, priv->token);
+        break;
+      case PROP_SUPERSEDES:
+        g_value_set_string (value, priv->supersedes_token);
+        break;
       default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
         break;
@@ -118,10 +153,21 @@ tpl_text_event_set_property (GObject *object,
       case PROP_MESSAGE_TYPE:
         priv->message_type = g_value_get_uint (value);
         break;
+      case PROP_EDIT_TIMESTAMP:
+        priv->edit_timestamp = g_value_get_int64 (value);
+        break;
       case PROP_MESSAGE:
         g_assert (priv->message == NULL);
         priv->message = g_value_dup_string (value);
         break;
+      case PROP_TOKEN:
+        g_assert (priv->token == NULL);
+        priv->token = g_value_dup_string (value);
+        break;
+      case PROP_SUPERSEDES:
+        g_assert (priv->supersedes_token == NULL);
+        priv->supersedes_token = g_value_dup_string (value);
+        break;
       default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
         break;
@@ -147,6 +193,7 @@ static void tpl_text_event_class_init (TplTextEventClass *klass)
   TplEventClass *event_class = TPL_EVENT_CLASS (klass);
   GParamSpec *param_spec;
 
+  object_class->dispose = tpl_text_event_dispose;
   object_class->finalize = tpl_text_event_finalize;
   object_class->get_property = tpl_text_event_get_property;
   object_class->set_property = tpl_text_event_set_property;
@@ -160,6 +207,14 @@ static void tpl_text_event_class_init (TplTextEventClass *klass)
       G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
   g_object_class_install_property (object_class, PROP_MESSAGE_TYPE, param_spec);
 
+  param_spec = g_param_spec_int64 ("edit-timestamp",
+      "Timestamp of edit message",
+      "message-{sent,received} if this is an edit, or 0 otherwise.",
+      G_MININT64, G_MAXINT64, 0,
+      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_EDIT_TIMESTAMP,
+      param_spec);
+
   param_spec = g_param_spec_string ("message",
       "Message",
       "The text message of the log event",
@@ -167,6 +222,20 @@ static void tpl_text_event_class_init (TplTextEventClass *klass)
       G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
   g_object_class_install_property (object_class, PROP_MESSAGE, param_spec);
 
+  param_spec = g_param_spec_string ("message-token",
+      "Message Token",
+      "The message-token field of this message.",
+      NULL,
+      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_TOKEN, param_spec);
+
+  param_spec = g_param_spec_string ("supersedes-token",
+      "Message Token",
+      "The message-token field of the message that this one supersedes.",
+      NULL,
+      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_SUPERSEDES, param_spec);
+
   g_type_class_add_private (object_class, sizeof (TplTextEventPriv));
 }
 
@@ -235,6 +304,76 @@ tpl_text_event_get_message (TplTextEvent *self)
 }
 
 
+/**
+ * tpl_text_event_get_message_token
+ * @self: a #TplTextEvent
+ *
+ * Returns: the same message as the #TplTextEvent:message-token property
+ */
+const gchar *
+tpl_text_event_get_message_token (TplTextEvent *self)
+{
+  g_return_val_if_fail (TPL_IS_TEXT_EVENT (self), NULL);
+
+  return self->priv->token;
+}
+
+
+/**
+ * tpl_text_event_get_supersedes_token
+ * @self: a #TplTextEvent
+ *
+ * Returns: the same message as the #TplTextEvent:supersedes-token property
+ */
+const gchar *
+tpl_text_event_get_supersedes_token (TplTextEvent *self)
+{
+  g_return_val_if_fail (TPL_IS_TEXT_EVENT (self), NULL);
+
+  return self->priv->supersedes_token;
+}
+
+
+/**
+ * _tpl_text_event_add_supersedes
+ * @self: a #TplTextEvent
+ * @old_event: (transfer none): an #TplTextEvent which this one supersedes
+ *
+ * If there are other known entries in the message edit/succession chain,
+ * they should be added to old_event before linking these two events,
+ * as they will be copied onto this event for convenience.
+ */
+void
+_tpl_text_event_add_supersedes (TplTextEvent *self,
+    TplTextEvent *old_event)
+{
+  GList *l;
+
+  g_object_ref (old_event);
+  g_queue_push_tail (&self->priv->supersedes, old_event);
+
+  for (l = old_event->priv->supersedes.head; l != NULL; l = g_list_next (l))
+    g_queue_push_tail (&self->priv->supersedes, g_object_ref (l->data));
+
+  if (self->priv->supersedes_token == NULL)
+    self->priv->supersedes_token = g_strdup (old_event->priv->token);
+}
+
+
+/**
+ * tpl_text_event_get_supersedes
+ * @self: a #TplTextEvent
+ *
+ * Returns: (transfer none): A #GList of #TplTextEvent that this event
+ * supersedes.
+ */
+GList *
+tpl_text_event_get_supersedes (TplTextEvent *self)
+{
+  return self->priv->supersedes.head;
+}
+
+
 /**
  * tpl_text_event_get_message_type
  * @self: a #TplTextEvent
@@ -251,3 +390,18 @@ tpl_text_event_get_message_type (TplTextEvent *self)
 }
 
 
+/**
+ * tpl_text_event_get_edit_timestamp
+ * @self: a #TplTextEvent
+ *
+ * Returns: the same value as the #TplTextEvent:edit-timestamp property
+ */
+gint64
+tpl_text_event_get_edit_timestamp (TplTextEvent *self)
+{
+  g_return_val_if_fail (TPL_IS_TEXT_EVENT (self), 0);
+
+  return self->priv->edit_timestamp;
+}
+
+
